using Godot;

public class HeatSystem
{
    private readonly TileGrid _tiles;
    private readonly float _diffusionCoefficient;

    private float[] _nextTemperatures = null;

    public HeatSystem(TileGrid tiles, float diffusionCoefficient)
    {
        _tiles = tiles;
        _diffusionCoefficient = diffusionCoefficient;
        _nextTemperatures = new float[tiles.Count];
    }

    public void Process()
    {
        GenerateHeat();
        Diffuse();
    }


    private float TransferHeat(float t0, float alpha, float nx, float ny, float px, float py)
    {
        float d2tdx2 = nx - 2 * t0 + px;
        float d2tdy2 = ny - 2 * t0 + py;
        return t0 + alpha * (d2tdx2 + d2tdy2);
    }

    private void GetNeighbourTemperatures(int x, int y, out float nx, out float ny, out float px, out float py)
    {
        // default value
        var t = _tiles[x, y].temperature;

        nx = x > 0 ? _tiles[x - 1, y].temperature : t;
        px = x < _tiles.Size - 1 ? _tiles[x + 1, y].temperature : t;

        ny = y > 0 ? _tiles[x, y - 1].temperature : t;
        py = y < _tiles.Size - 1 ? _tiles[x, y + 1].temperature : t;
    }
    private void GenerateHeat()
    {
        for (int i = 0; i < _tiles.Count; i++)
        {
            var tile = _tiles[i];
            var type = tile.type;
            tile.temperature += type.HeatGeneration;
            ApplyHeatDamage(ref tile);
            _tiles[i] = tile;
        }
    }

    private void ApplyHeatDamage(ref Tile tile)
    {
        if (!(tile.type is IDamageable damageable))
            return;

        var surplus = tile.temperature - damageable.Threshold;
        if (surplus < 0)
            return;

        // TODO: parameterised balancing
        tile.currentHealth -= surplus;
        tile.currentHealth = Mathf.Clamp(tile.currentHealth, 0, 1);
    }

    private void Diffuse()
    {
        var D = _diffusionCoefficient;

        for (int i = 0; i < _tiles.Count; i++)
        {
            float t = _tiles[i].temperature;

            _tiles.MapIndex(i, out var x, out var y);

            GetNeighbourTemperatures(
                x, y,
                out var nx,
                out var ny,
                out var px,
                out var py);

            var temperature = TransferHeat(t, D, nx, ny, px, py);

            // TODO: what if it's really really cold out?
            temperature = Mathf.Max(0, temperature);
            _nextTemperatures[i] = temperature;
        }

        for (int i = 0; i < _tiles.Count; i++)
        {
            var tile = _tiles[i];
            tile.temperature = _nextTemperatures[i];
            _tiles[i] = tile;
        }
    }

}
