using Godot;
using System;

public class WorldGrid : Node2D
{
    [Export]
    private PackedScene TileScene { get; set; }

    [Export]
    private Resource _tileGridResource;
    private TileGrid _tileGrid;

    [Export]
    public int CellSize { get; set; }

    [Export]
    public float DiffusionCoefficient { get; set; } = 0.01f;

    public struct TileView
    {
        public ShaderMaterial material;
    }
    private TileView[] _tileViews;
    private HeatSystem _heat;

    // Called when the node enters the scene tree for the first time.
    public override void _Ready()
    {
        _tileGrid = (TileGrid)_tileGridResource;

        GenerateCanvasItems(_tileGrid);

        _heat = new HeatSystem(_tileGrid, DiffusionCoefficient);
    }

    #region Positioning
    public void GetGridPos(Vector2 position, out int x, out int y)
    {
        x = Mathf.FloorToInt(position.x / CellSize);
        y = Mathf.FloorToInt(position.y / CellSize);
    }

    public bool IsInBounds(int x, int y) => _tileGrid.IsInBounds(x, y);
    #endregion

    #region Interaction
    public void SetTileType(int x, int y, TileType tileType)
    {
        if (!_tileGrid.IsInBounds(x, y))
            return;

        _tileGrid.MapPosition(x, y, out var idx);

        var tile = _tileGrid[idx];
        tile.type = tileType;
        _tileGrid[x, y] = tile;
    }
    #endregion

    private void Clear()
    {
        int children = this.GetChildCount();
        GD.Print(children);
        for (int i = 0; i < children; i++)
        {
            var child = this.GetChild(i);
            child.QueueFree();
        }

        _tileViews = null;
    }

    private void GenerateCanvasItems(TileGrid grid)
    {
        this.Clear();

        _tileViews = new TileView[grid.Count];

        for (int i = 0; i < grid.Count; i++)
        {
            _tileGrid.MapIndex(i, out var x, out var y);

            var node = TileScene.Instance<Node2D>();
            this.AddChild(node);
            var position = new Vector2
            {
                x = (x + .5f) * CellSize,
                y = (y + .5f) * CellSize
            };
            node.Position = position;
            var canvasItem = (CanvasItem)node;

            var tile = _tileGrid[i];

            TileView view = default;
            view.material = (ShaderMaterial)canvasItem.Material;
            _tileViews[i] = view;
        }
    }

    public ShaderMaterial GetTileMaterial(int idx) => _tileViews[idx].material;

    public void _on_Clock_OnTick(int ticks) => _heat.Process();
}
